package net.handle.server.storage;

import com.mongodb.ErrorCategory;
import com.mongodb.MongoException;
import com.mongodb.MongoWriteException;
import com.mongodb.client.*;
import com.mongodb.client.model.CountOptions;
import com.mongodb.client.model.Filters;
import com.mongodb.client.result.DeleteResult;
import com.mongodb.client.result.UpdateResult;

import net.cnri.util.StreamTable;
import net.handle.hdllib.*;
import org.bson.Document;
import org.bson.conversions.Bson;

import java.io.Closeable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

/**
 * A handle storage class for connecting to MongoDB.
 * <p>
 * To use this storage module configure your config.dct file with the following options.
 * <pre>
 * "storage_type" = "CUSTOM"
 * "storage_class" = "net.handle.server.storage.MongoDbHandleStorage"
 * "storage_config" = {
 *     "connectionUri" = "mongodb://localhost:27017"
 *     "databaseName" = "handle"
 *     "handlesCollectionName" = "handles"
 *     "nasCollectionName" = "nas"
 *     "maxTimeMs" = "30000"
 *     "maxTimeMsLongRunning" = "3600000"
 * }
 * </pre>
 * The values given in the above example are the default values; "storage_config" can be omitted
 * if no changes are desired.
 */
public class MongoDbHandleStorage implements HandleStorage {

    private static final String HANDLE_DB = "handle";
    private static final long DEFAULT_MAX_TIME_MS = 30_000; // 30 seconds
    private static final long DEFAULT_MAX_TIME_MS_LONG_RUNNING = 1000 * 60 * 60 ; // 1 hour

    private MongoCollection<Document> nasCollection;
    private MongoCollection<Document> handlesCollection;
    private MongoClient client;
    private long maxTimeMs;
    private long maxTimeMsLongRunning;
    private boolean isReadOnly;

    @Override
    public void init(StreamTable config) throws Exception {
        isReadOnly = config.getBoolean("isReadOnly");
        maxTimeMs = config.getLong("maxTimeMs", DEFAULT_MAX_TIME_MS);
        maxTimeMsLongRunning = config.getLong("maxTimeMsLongRunning", DEFAULT_MAX_TIME_MS_LONG_RUNNING);

        String connectionUri = config.getStr("connectionUri", "mongodb://localhost:27017");
        client = MongoClients.create(connectionUri);

        String databaseName = config.getStr("databaseName", HANDLE_DB);
        MongoDatabase db = client.getDatabase(databaseName);

        nasCollection = db.getCollection(config.getStr("nasCollectionName", "nas"));
        handlesCollection = db.getCollection(config.getStr("handlesCollectionName", "handles"));
    }

    @Override
    public boolean haveNA(byte[] authHandle) throws HandleException {
        authHandle = Util.upperCase(authHandle);
        String prefix = Util.decodeString(authHandle);
        try {
            long count = nasCollection.countDocuments(new Document("_id", prefix), new CountOptions().limit(1).maxTime(maxTimeMs, TimeUnit.MILLISECONDS));
            return count != 0;
        } catch (MongoException me) {
            throw new HandleException(HandleException.INTERNAL_ERROR, "Error in haveNA", me);
        }
    }

    @Override
    public void setHaveNA(byte[] authHandle, boolean flag) throws HandleException {
        if (isReadOnly) {
            throw new HandleException(HandleException.STORAGE_RDONLY, "Storage is read-only");
        }
        byte[] handle = Util.upperCase(authHandle);
        String prefix = Util.decodeString(handle);
        try {
            if (flag) {
                nasCollection.insertOne(new Document("_id", prefix));
            } else {
                nasCollection.deleteOne(new Document("_id", prefix));
            }
        } catch (MongoException me) {
            throw new HandleException(HandleException.INTERNAL_ERROR, "Error recording NA data", me);
        }
    }

    @Override
    public void createHandle(byte[] handleBytes, HandleValue[] values) throws HandleException {
        if (isReadOnly) {
            throw new HandleException(HandleException.STORAGE_RDONLY, "Storage is read-only");
        }
        String handle = Util.decodeString(handleBytes);
        Document document = documentOfHandleRecord(handle, values);
        try {
            handlesCollection.insertOne(document);
        } catch (MongoWriteException me) {
            if (me.getError().getCategory() == ErrorCategory.DUPLICATE_KEY) {
                throw new HandleException(HandleException.HANDLE_ALREADY_EXISTS, handle, me);
            } else {
                throw new HandleException(HandleException.INTERNAL_ERROR, "Error creating handle", me);
            }
        } catch (MongoException me) {
            throw new HandleException(HandleException.INTERNAL_ERROR, "Error creating handle", me);
        }
    }

    private Document documentOfHandleRecord(String handle, HandleValue[] values) {
        Document doc = new Document("_id", handle);
        doc.put("values", Arrays.stream(values)
            .map(value -> Document.parse(GsonUtility.getGson().toJson(value)))
            .collect(Collectors.toList()));
        return doc;
    }

    @Override
    public boolean deleteHandle(byte[] handleBytes) throws HandleException {
        if (isReadOnly) {
            throw new HandleException(HandleException.STORAGE_RDONLY, "Storage is read-only");
        }
        String handle = Util.decodeString(handleBytes);
        try {
            DeleteResult result = handlesCollection.deleteOne(new Document("_id", handle));
            return result.getDeletedCount() > 0;
        } catch (MongoException me) {
            throw new HandleException(HandleException.INTERNAL_ERROR, "Error deleting handle", me);
        }
    }

    @Override
    public byte[][] getRawHandleValues(byte[] handleBytes, int[] indexList, byte[][] typeList) throws HandleException {
        String handle = Util.decodeString(handleBytes);
        Document query = new Document("_id", handle);
        Document document;
        try {
            document = handlesCollection.find(query).limit(1).maxTime(maxTimeMs, TimeUnit.MILLISECONDS).first();
        } catch (Exception me) {
            throw new HandleException(HandleException.INTERNAL_ERROR, "Error in getRawHandleValues", me);
        }
        if (document == null) {
            return null;
        }
        List<HandleValue> allValues = document.getList("values", Document.class).stream()
            .map(doc -> GsonUtility.getGson().fromJson(doc.toJson(), HandleValue.class))
            .collect(Collectors.toList());
        List<HandleValue> filteredValues;
        boolean includeAllValues = (indexList == null || indexList.length == 0) && (typeList == null  || typeList.length == 0);
        if (includeAllValues) {
            filteredValues = allValues;
        } else {
            filteredValues = new ArrayList<>();
            for (HandleValue value : allValues) {
                int index = value.getIndex();
                byte[] type = value.getType();
                if (Util.isParentTypeInArray(typeList, type) || Util.isInArray(indexList, index)) {
                    filteredValues.add(value);
                }
            }
        }
        byte[][] valBytes = Encoder.encodeHandleValues(filteredValues.toArray(new HandleValue[0]));
        return valBytes;
    }

    @Override
    public void updateValue(byte[] handleBytes, HandleValue[] values) throws HandleException {
        if (isReadOnly) {
            throw new HandleException(HandleException.STORAGE_RDONLY, "Storage is read-only");
        }
        String handle = Util.decodeString(handleBytes);
        Document document = documentOfHandleRecord(handle, values);
        UpdateResult result;
        try {
            result = handlesCollection.replaceOne(new Document("_id", handle), document);
        } catch (MongoException me) {
            throw new HandleException(HandleException.INTERNAL_ERROR, "Error in updateValue", me);
        }
        if (result.getMatchedCount() == 0) {
            throw new HandleException(HandleException.HANDLE_DOES_NOT_EXIST, "Cannot modify non-existent handle");
        }
    }

    @Override
    public void scanHandles(ScanCallback callback) throws HandleException {
        try (MongoCursor<Document> documents = handlesCollection.find().projection(new Document("_id", 1)).maxTime(maxTimeMsLongRunning, TimeUnit.MILLISECONDS).iterator()) {
            while (documents.hasNext()) {
                Document document = documents.next();
                String handle = document.getString("_id");
                byte[] handleBytes = Util.encodeString(handle);
                callback.scanHandle(handleBytes);
            }
        } catch (HandleException he) {
            throw he;
        } catch (MongoException me) {
            throw new HandleException(HandleException.INTERNAL_ERROR, "Error in scanHandles", me);
        }
    }

    @Override
    public void scanNAs(ScanCallback callback) throws HandleException {
        try (MongoCursor<Document> documents = nasCollection.find().projection(new Document("_id", 1)).maxTime(maxTimeMsLongRunning, TimeUnit.MILLISECONDS).iterator()) {
            while (documents.hasNext()) {
                Document document = documents.next();
                String prefix = document.getString("_id");
                byte[] prefixBytes = Util.encodeString(prefix);
                callback.scanHandle(prefixBytes);
            }
        } catch (HandleException he) {
            throw he;
        } catch (MongoException me) {
            throw new HandleException(HandleException.INTERNAL_ERROR, "Error in scanHandles", me);
        }
    }

    @Override
    public Enumeration<byte[]> getHandlesForNA(byte[] naHdl) throws HandleException {
        boolean isZeroNA = Util.startsWithCI(naHdl, Common.NA_HANDLE_PREFIX);
        if (isZeroNA) naHdl = Util.getSuffixPart(naHdl);
        return new HdlsForNAEnum(naHdl);
    }

    @Override
    public void deleteAllRecords() throws HandleException {
        if (isReadOnly) {
            throw new HandleException(HandleException.STORAGE_RDONLY, "Storage is read-only");
        }
        try {
            handlesCollection.drop();
            nasCollection.drop();
        } catch (MongoException me) {
            throw new HandleException(HandleException.INTERNAL_ERROR, "Error deleting all records", me);
        }
    }

    @Override
    public void checkpointDatabase() throws HandleException {
        throw new HandleException(HandleException.SERVER_ERROR, "Checkpoint not supported in this storage type");
    }

    @Override
    public void shutdown() {
        if(client != null) {
            try {
                client.close();
            } catch (Exception e) {
                // no-op
            }
        }
    }

    private class HdlsForNAEnum implements Enumeration<byte[]>, Closeable {

        private MongoCursor<Document> cursor;

        HdlsForNAEnum(byte[] prefixBytes) throws HandleException {
            String prefix = Util.decodeString(prefixBytes);
            boolean listingDerivedPrefixes = Util.startsWithCI(prefixBytes, Common.NA_HANDLE_PREFIX);
            if (listingDerivedPrefixes) {
                prefix += ".";
            } else {
                prefix += "/";
            }
            Pattern searchPattern = Pattern.compile("^" + Pattern.quote(prefix), Pattern.CASE_INSENSITIVE);
            Bson query = Filters.regex("_id", searchPattern);
            try {
                cursor = handlesCollection.find(query).projection(new Document("_id", 1)).maxTime(maxTimeMsLongRunning, TimeUnit.MILLISECONDS).iterator();
            } catch (MongoException e) {
                throw new HandleException(HandleException.INTERNAL_ERROR, "Error listing handles", e);
            }
        }

        @Override
        public boolean hasMoreElements() {
            return cursor.hasNext();
        }

        @Override
        public byte[] nextElement() {
            Document document = cursor.next();
            String handle = document.getString("_id");
            return Util.encodeString(handle);
        }

        @Override
        public void close() {
            if (cursor != null) {
                cursor.close();
            }
        }
    }
}
